def get_user_data_dir(create: bool = False) -> str: ...
def _set_context_data_dir() -> None: ...
def _set_context_ca_bundle_path(ca_bundle_path: str) -> None: ...
def _set_context_network_enabled() -> None: ...
def set_use_global_context(active: bool | None = None) -> None: ...
def _clear_proj_error() -> None: ...
def _get_proj_error() -> str: ...
